# Copyright DataStax, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

import sys
from typing import cast


def get_python_version() -> tuple[int, int, int]:
    """
    Get the version of the currently-running Python interpreter.

    Returns:
        version: (major, minor, revision) as a tuple[int, int, int]
    """

    return cast(tuple[int, int, int], tuple(sys.version_info[:3]))
